# -*- coding: utf-8 -*-
# author:chao.yy
# email:yuyc@ishangqi.com
# date:2021/11/24 7:34 下午
# Copyright (C) 2021 The lesscode Team

from tornado.options import options, define

from lesscode.db.connection_info import ConnectionInfo
from lesscode.web.business_exception import BusinessException
from lesscode.web.status_code import StatusCode

define("database", default={}, type=dict, help="数据库链接资源池,内部使用")
define("conn_info", default=None, type=list, help="数据库链接信息")


class InitConnectionPool:
    """
    初始化连接池
    """

    @staticmethod
    async def create_pool():
        """
        获取配置的链接信息，生成对应的连接池
        :return:
        """
        if options.conn_info:
            for conn_info in options.conn_info:
                if conn_info.enable:
                    if options.database.keys().__contains__(conn_info.name):
                        # 相同key的连接池已经存在，抛出异常
                        raise BusinessException(StatusCode.RESOURCE_EXIST(f"{conn_info.name}连接池"))
                    pool = None
                    # 依据数据库类型进行链接池创建
                    if conn_info.dialect == "postgresql":
                        pool = await InitConnectionPool.create_pool_postgresql(conn_info)
                    elif conn_info.dialect == "mysql":
                        pool = await InitConnectionPool.create_pool_mysql(conn_info)
                    elif conn_info.dialect == "mongodb":
                        pool = InitConnectionPool.create_pool_mongodb(conn_info)
                    elif conn_info.dialect == "elasticsearch":
                        pool = await InitConnectionPool.create_pool_es(conn_info)
                    if pool:
                        options.database[conn_info.name] = (pool, conn_info.dialect)

    @staticmethod
    async def create_pool_postgresql(conn_info: ConnectionInfo):
        """
        创建postgresql 异步连接池
        :param conn_info: 连接信息
        :return: AsyncConnectionPool
        """
        if conn_info.async_enable:
            from aiopg import create_pool
            pool = await create_pool(host=conn_info.host, port=conn_info.port, user=conn_info.user,
                                     password=conn_info.password,
                                     database=conn_info.db_name)
            return pool
        else:
            raise NotImplementedError

    @staticmethod
    async def create_pool_mysql(conn_info: ConnectionInfo):
        """
        创建mysql 异步连接池
        :param conn_info: 连接信息
        :return:
        """
        if conn_info.async_enable:
            import aiomysql
            pool = await aiomysql.create_pool(host=conn_info.host, port=conn_info.port, user=conn_info.user,
                                              password=conn_info.password,
                                              db=conn_info.db_name, autocommit=True, minsize=conn_info.min_size,
                                              maxsize=conn_info.max_size)
            return pool
        else:
            raise NotImplementedError

    @staticmethod
    def create_pool_mongodb(conn_info: ConnectionInfo):
        """
        创建mongodb 异步连接池
        :param conn_info: 连接信息
        :return:
        """
        if conn_info.async_enable:
            import motor
            conn_info_string = f"mongodb://{conn_info.user}:{conn_info.password}@{conn_info.host}:{conn_info.port}"
            if conn_info.params:
                if conn_info.params == "LDAP":
                    conn_info_string += "/?authMechanism=PLAIN"
                elif conn_info.params == "Password":
                    conn_info_string += "/?authSource=admin"
                elif conn_info.params == "X509":
                    conn_info_string += "/?authMechanism=MONGODB-X509"
            pool = motor.motor_tornado.MotorClient(conn_info_string)
            return pool
        else:
            raise NotImplementedError

    @staticmethod
    async def create_pool_es(conn_info: ConnectionInfo):
        """
        创建elasticsearch 异步连接池
        :param conn_info: 连接信息
        :return:
        """
        if conn_info.async_enable:
            from elasticsearch import AsyncElasticsearch
            host_str = conn_info.host.split(",")
            hosts = [f"{conn_info.user}:{conn_info.password}@{host}:{conn_info.port}" for host in host_str]
            pool = AsyncElasticsearch(hosts=hosts)
            return pool
        else:
            raise NotImplementedError



# -*- coding: utf-8 -*-
# author:chao.yy
# email:yuyc@ishangqi.com
# date:2021/11/4 11:00 上午
# Copyright (C) 2021 The lesscode Team

import threading

from tornado.web import RequestHandler

from lesscode.web.business_exception import BusinessException
from lesscode.web.status_code import StatusCode


class RouterMapping(list):
    """
    RouterMapping 类用于存储访问路径"URL"与处理器之间对应关系集合
    直接继承list
    """

    _instance_lock = threading.Lock()

    def __init__(self):
        super(RouterMapping, self).__init__()
        # 存放类名、处理方法、url 的元祖集合 url为方法上指定路径
        self.dynamicMethods = []
        # 存放url与具体业务处理方法的映射关系（url，function）
        self.handlerMapping = []

    @classmethod
    def instance(cls):
        if not hasattr(RouterMapping, "_instance"):
            with RouterMapping._instance_lock:
                if not hasattr(RouterMapping, "_instance"):
                    RouterMapping._instance = RouterMapping()
        return RouterMapping._instance


def Handler(url: str):
    """
    RequestHandler对应路径注册装饰器，完成处理类与url对应注册。
    :param url:
    :return:
    """

    def wrapper(cls):
        # 验证是否为RequestHandler 子类，仅注册其子类
        if not issubclass(cls, RequestHandler):
            raise RuntimeError("Handler注释器只能装饰在RequestHandler子类上")
        # 通过类名查找对应该类下的所有注册方法信息
        res = [item for item in RouterMapping.instance().dynamicMethods
               if item[0] == cls.__name__]
        for item in res:
            # 处理类 RequestHandler 子类 使用时统一继承BaseHandler
            handler = item[1]
            # 全路径 Handler+Mapping 组合
            full_url = url + item[2]
            # 判断是否存在重复注册情况，重复情况直接抛出异常
            if [router for router in RouterMapping.instance() if full_url in router]:
                raise BusinessException(StatusCode.RESOURCE_EXIST(f'路由"{full_url}"'))
            # 存储URL与 RequestHandler 子类对应关系，用于提供给Tornado注册使用
            RouterMapping.instance().append((full_url, cls))
            # 存储URL与处理方法的对应关系，用于调用分发使用
            RouterMapping.instance().handlerMapping.append((full_url, handler))

    return wrapper


def GetMapping(url: str):
    """
    用于类名、处理方法、url 的元祖集合注册处理，暂时GetMapping与PostMapping实现保持一致，预留入口为后期扩展提供支持。
    :param url:
    :return:
    """

    def wrapper(func):
        # 组合  类名、处理方法、url 的元祖 （url仅代表方法上Mapping装饰器的参数）
        RouterMapping.instance().dynamicMethods.append((func.__qualname__.replace('.' + func.__name__, ''), func, url))

    return wrapper


def PostMapping(url: str):
    """
    用于类名、处理方法、url 的元祖集合注册处理，暂时GetMapping与PostMapping实现保持一致，预留入口为后期扩展提供支持。
    :param url:
    :return:
    """

    def wrapper(func):
        # 组合  类名、处理方法、url 的元祖 （url仅代表方法上Mapping装饰器的参数）
        RouterMapping.instance().dynamicMethods.append((func.__qualname__.replace('.' + func.__name__, ''), func, url))

    return wrapper



# -*- coding: utf-8 -*-
# author:chao.yy
# email:yuyc@ishangqi.com
# date:2021/11/30 3:27 下午
# Copyright (C) 2021 The lesscode Team
import abc
import logging

from tornado.options import define, options

from lesscode.db.condition_wrapper import ConditionWrapper
from lesscode.web.business_exception import BusinessException
from lesscode.web.status_code import StatusCode

define("echo_sql", default=True, type=bool, help="是否启用SQL输出")
define("max_limit", default=500, type=int, help="单次查询最大数量")


class BaseSqlHelper(abc.ABC):
    """
    BaseSqlHelper 是数据操作的基础抽象类，用于定义操作标准接口
    """

    def __init__(self, pool):
        """
        初始化sql工具
        :param pool: 连接池对象
        """
        self.pool = pool

    @abc.abstractmethod
    async def insert_data(self, table_name: str, data):
        """
        对于插入接口的整合，不需要考虑是单条插入还是多条插入，自动区分
        :param table_name:
        :param data:
        :return:
        """
        pass

    @abc.abstractmethod
    async def insert_one_data(self, table_name: str, data: dict):
        """
        新增记录插入
        :param table_name: 表名
        :param data: 待插入数据
        :return:
        """
        pass

    @abc.abstractmethod
    async def insert_many_data(self, table_name: str, data: list):
        """
        批量插入
        :param table_name:
        :param data:
        :return:
        """
        pass

    @abc.abstractmethod
    async def update_data(self, condition_wrapper: ConditionWrapper, param: dict):
        """
        按条件更新
        :param condition_wrapper:
        :param param:
        :return:
        """
        pass

    @abc.abstractmethod
    async def delete_data(self, condition_wrapper: ConditionWrapper):
        """
        按条件删除
        :param condition_wrapper: 删除条件
        :return:
        """
        pass

    @abc.abstractmethod
    async def fetchone_data(self, condition_wrapper: ConditionWrapper):
        """
        查询单条数据
        :param condition_wrapper: 查询条件信息
        :return:
        """
        pass

    @abc.abstractmethod
    async def fetchall_data(self, condition_wrapper: ConditionWrapper):
        """
        查询多条数据
        :param condition_wrapper: 查询条件信息
        :return:
        """
        pass

    @abc.abstractmethod
    async def fetchall_page(self, condition_wrapper: ConditionWrapper, page_num=1, page_size=10):
        """
        分页查询多条数据
        :param condition_wrapper: 查询条件信息，不含分页信息
        :param page_num: 当前页码
        :param page_size: 每页数量
        :return:
        """
        pass

    @abc.abstractmethod
    async def execute_sql(self, sql: str, param=None):
        """
        执行sql 仅返回影响数量
        :param sql:
        :param param:
        :return:
        """
        pass

    @abc.abstractmethod
    async def executemany_sql(self, sql: str, param=None):
        """
        执行批量sql 仅返回影响数量
        :param sql:
        :param param:
        :return:
        """
        pass

    @abc.abstractmethod
    async def execute_fetchone(self, sql: str, param=None):
        """
        查询单条数据
        :param sql: 待执行的Sql语句
        :param param: 参数
        :return:
        """
        pass

    @abc.abstractmethod
    async def execute_fetchall(self, sql: str, param=None):
        """
        查询多条数据
        :param sql: 待执行的Sql语句
        :param param: 参数
        :return:
        """
        pass

    @abc.abstractmethod
    def prepare_insert_sql(self, table_name: str, item: dict):
        """
        组装插入sql
        :param table_name:
        :param item:
        :return:
        """
        pass

    @abc.abstractmethod
    def prepare_update_sql(self, condition_wrapper: ConditionWrapper, param: dict):
        """
        组装更新sql
        :param condition_wrapper:
        :param param:
        :return:
        """
        pass

    @abc.abstractmethod
    def prepare_delete_sql(self, condition_wrapper: ConditionWrapper):
        """
        组装删除SQl
        :param condition_wrapper:
        :return:
        """
        pass

    @abc.abstractmethod
    def prepare_condition_sql(self, conditions: list):
        """
        拼接条件SQL 搜集参数信息
        :param conditions: 条件信息
        :return: 条件sql片段, values 对应参数
        """
        pass

    @abc.abstractmethod
    def prepare_query_sql(self, condition_wrapper: ConditionWrapper):
        """
        获取完整拼接查询SQL
        :param condition_wrapper: 查询条件对象
        :return:sql 查询sql, values 对应参数
        """
        pass

    @abc.abstractmethod
    def prepare_page_sql(self, condition_wrapper: ConditionWrapper, page_num: int, page_size: int):
        """
        获取分页SQL
        :param condition_wrapper: 查询参数
        :param page_num: 当前页
        :param page_size: 每页记录数
        :return:page_sql 分页sql, values 对应参数, count_sql 统计总数sql
        """
        pass

    @staticmethod
    def dict_row(desc, values):
        """
        记录解析方法
        :param desc:
        :param values:
        :return:
        """
        if desc is None:
            raise BusinessException(StatusCode.INTERNAL_SERVER_ERROR("没有需要解析的结果集"))
        titles = [c[0] for c in desc]
        if isinstance(values, list):
            return [dict(zip(titles, value)) for value in values]
        elif isinstance(values, tuple):
            return dict(zip(titles, values))


def echo_sql(func):
    """
    输出执行SQL的装饰器，要求方法 第一个是SQL 第二个是参数
    :param func:
    :return:
    """

    def wrapper(self, *args, **kwargs):
        if options.echo_sql:
            if len(args) > 1:
                logging.info(f"SQL：{str(args[0])}；参数：{str(args[1:])}")
            else:
                logging.info(f"SQL：{str(args[0])}")
        return func(self, *args, **kwargs)

    return wrapper
